# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fetch a build artifact (ported from google3 with
minor modifications, especially without any google3 specific library
dependencies)."""

import io
import json
import os
import re

import apiclient
from oauth2client.service_account import ServiceAccountCredentials

from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.metrics import logs

# 20 MB default chunk size.
DEFAULT_CHUNK_SIZE = 20 * 1024 * 1024

# Maximum number of retries for artifact access.
MAX_RETRIES = 20


def execute_request_with_retries(request):
  """Executes request and retries on failure."""
  result = None
  for _ in range(MAX_RETRIES):
    try:
      result = request.execute()
      break
    except Exception:
      pass

  return result


def download_artifact(client, bid, target, attempt_id, name, output_directory,
                      output_filename):
  """Download one artifact."""
  artifact_query = client.buildartifact().get(
      buildId=bid, target=target, attemptId=attempt_id, resourceId=name)
  artifact = execute_request_with_retries(artifact_query)
  if artifact is None:
    logs.log_error('No artifact found with name %s, target %s and build id %s.'
                   % (name, target, bid))
    return False

  # Lucky us, we always have the size.
  size = int(artifact['size'])

  chunksize = -1
  if size >= DEFAULT_CHUNK_SIZE:
    chunksize = DEFAULT_CHUNK_SIZE

  # Just like get, except get_media.
  dl_request = client.buildartifact().get_media(
      buildId=bid, target=target, attemptId=attempt_id, resourceId=name)

  if output_filename:
    file_name = output_filename
  else:
    file_name = name.replace('signed/', '')

  output_path = os.path.join(output_directory, file_name)
  # If the artifact already exists, then bail out.
  if os.path.exists(output_path) and os.path.getsize(output_path) == size:
    logs.log('Artifact %s already exists, skipping download.' % name)
    return output_path

  logs.log('Downloading artifact %s.' % name)
  with io.FileIO(output_path, mode='wb') as file_handle:
    downloader = apiclient.http.MediaIoBaseDownload(
        file_handle, dl_request, chunksize=chunksize)
    done = False

    while not done:
      status, done = downloader.next_chunk()
      if status:
        size_completed = int(status.resumable_progress)
        percent_completed = (size_completed * 100.0) / size
        logs.log('%.1f%% complete.' % percent_completed)

  return output_path


def get_artifacts_for_build(client, bid, target, attempt_id='latest'):
  """Return list of artifacts for a given build."""
  request = client.buildartifact().list(
      buildId=bid, target=target, attemptId=attempt_id)

  artifacts = []
  while request:
    result = execute_request_with_retries(request)
    if not result:
      break
    if result and 'artifacts' in result:
      for artifact in result['artifacts']:
        artifacts.append(artifact)
    request = client.buildartifact().list_next(request, result)

  return artifacts


def get_client():
  """Return client with connection to build apiary."""
  # Connect using build apiary service account credentials.
  build_apiary_service_account_private_key = db_config.get_value(
      'build_apiary_service_account_private_key')
  if not build_apiary_service_account_private_key:
    logs.log(
        'Android build apiary credentials are not set, skip artifact fetch.')
    return None

  credentials = ServiceAccountCredentials.from_json_keyfile_dict(
      json.loads(build_apiary_service_account_private_key),
      scopes='https://www.googleapis.com/auth/androidbuild.internal')
  client = apiclient.discovery.build(
      'androidbuildinternal',
      'v2beta1',
      credentials=credentials,
      cache_discovery=False)

  return client


def get_latest_artifact_info(branch, target, signed=False):
  """Return latest artifact for a branch and target."""
  client = get_client()
  if not client:
    return None

  request = client.build().list(
      buildType='submitted',
      branch=branch,
      target=target,
      successful=True,
      maxResults=1,
      signed=signed)
  builds = execute_request_with_retries(request)
  if not builds:
    logs.log_error(
        'No artifact found for target %s, branch %s.' % (target, branch))
    return None

  build = builds['builds'][0]
  bid = build['buildId']
  target = build['target']['name']
  return {'bid': bid, 'branch': branch, 'target': target}


def get(bid, target, regex, output_directory, output_filename=None):
  """Return artifact for a given build id, target and file regex."""
  client = get_client()
  if not client:
    return None

  # Run the script to fetch the artifact.
  return run_script(
      client=client,
      bid=bid,
      target=target,
      regex=regex,
      output_directory=output_directory,
      output_filename=output_filename)


def run_script(client, bid, target, regex, output_directory, output_filename):
  """Download artifacts as specified."""
  artifacts = get_artifacts_for_build(
      client=client, bid=bid, target=target, attempt_id='latest')
  if not artifacts:
    logs.log_error(
        'No artifact found for target %s, build id %s.' % (target, bid))
    return False

  regex = re.compile(regex)
  result = None
  for artifact in artifacts:
    artifact_name = artifact['name']

    # Exclude signature files.
    if artifact_name.endswith('.SIGN_INFO'):
      continue

    if regex.match(artifact_name):
      result = download_artifact(
          client=client,
          bid=bid,
          target=target,
          attempt_id='latest',
          name=artifact_name,
          output_directory=output_directory,
          output_filename=output_filename)

  return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for regression_task."""
# pylint: disable=unused-argument
# pylint: disable=protected-access

import os
import unittest

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.bot.tasks import regression_task
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class WriteToBigQueryTest(unittest.TestCase):
  """Test write_to_big_query."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.big_query.write_range',
    ])

    self.testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        fuzzer_name='libfuzzer',
        overridden_fuzzer_name='libfuzzer_pdf',
        job_type='some_job')

  def test_write(self):
    """Tests write."""
    regression_task.write_to_big_query(self.testcase, 456, 789)
    self.mock.write_range.assert_called_once_with(
        table_id='regressions',
        testcase=self.testcase,
        range_name='regression',
        start=456,
        end=789)


class TestcaseReproducesInRevisionTest(unittest.TestCase):
  """Test _testcase_reproduces_in_revision."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.setup_regular_build',
        'clusterfuzz._internal.bot.testcase_manager.test_for_crash_with_retries',
        'clusterfuzz._internal.bot.testcase_manager.check_for_bad_build',
    ])

  def test_error_on_failed_setup(self):
    """Ensure that we throw an exception if we fail to set up a build."""
    os.environ['APP_NAME'] = 'app_name'
    # No need to implement a fake setup_regular_build. Since it's doing nothing,
    # we won't have the build directory properly set.
    with self.assertRaises(errors.BuildSetupError):
      regression_task._testcase_reproduces_in_revision(
          None, '/tmp/blah', 'job_type', 1, should_log=False)


@test_utils.with_cloud_emulators('datastore')
class TestFoundRegressionNearExtremeRevisions(unittest.TestCase):
  """Test found_regression_near_extreme_revisions."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.regression_task.save_regression_range',
        'clusterfuzz._internal.bot.tasks.regression_task._testcase_reproduces_in_revision',
    ])

    # Keep a dummy test case. Values are not important, but we need an id.
    self.testcase = data_types.Testcase()
    self.testcase.put()

    self.revision_list = [1, 2, 5, 8, 9, 12, 15, 19, 21, 22]

  def test_near_max_revision(self):
    """Ensure that we return True if this is a very recent regression."""

    def testcase_reproduces(testcase,
                            testcase_file_path,
                            job_type,
                            revision,
                            should_log=True,
                            min_revision=None,
                            max_revision=None):
      return revision > 20

    self.mock._testcase_reproduces_in_revision.side_effect = testcase_reproduces

    regression_task.found_regression_near_extreme_revisions(
        self.testcase, '/a/b', 'job_name', self.revision_list, 0, 9)

  def test_at_min_revision(self):
    """Ensure that we return True if we reproduce in min revision."""
    self.mock._testcase_reproduces_in_revision.return_value = True

    regression_task.found_regression_near_extreme_revisions(
        self.testcase, '/a/b', 'job_name', self.revision_list, 0, 9)

  def test_not_at_extreme_revision(self):
    """Ensure that we return False if we didn't regress near an extreme."""

    def testcase_reproduces(testcase,
                            testcase_file_path,
                            job_type,
                            revision,
                            should_log=True,
                            min_revision=None,
                            max_revision=None):
      return revision > 10

    self.mock._testcase_reproduces_in_revision.side_effect = testcase_reproduces

    regression_task.found_regression_near_extreme_revisions(
        self.testcase, '/a/b', 'job_name', self.revision_list, 0, 9)


def _sample(input_list, count):
  """Helper function to deterministically sample a list."""
  assert count <= len(input_list)
  return input_list[:count]


@test_utils.with_cloud_emulators('datastore')
class ValidateRegressionRangeTest(unittest.TestCase):
  """Tests for validate_regression_range."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.regression_task._testcase_reproduces_in_revision',
        'random.sample',
    ])

    self.mock.sample.side_effect = _sample

  def test_no_earlier_revisions(self):
    """Make sure we don't throw exceptions if nothing is before min revision."""
    testcase = data_types.Testcase()
    testcase.put()

    self.mock._testcase_reproduces_in_revision.return_value = False
    result = regression_task.validate_regression_range(testcase, '/a/b',
                                                       'job_type', [0], 0)
    self.assertTrue(result)

  def test_one_earlier_revision(self):
    """Test a corner-case with few revisions earlier than min revision."""
    testcase = data_types.Testcase()
    testcase.put()

    self.mock._testcase_reproduces_in_revision.return_value = False
    result = regression_task.validate_regression_range(testcase, '/a/b',
                                                       'job_type', [0, 1, 2], 1)
    self.assertTrue(result)

  def test_invalid_range(self):
    """Ensure that we handle invalid ranges correctly."""
    testcase = data_types.Testcase()
    testcase.put()

    self.mock._testcase_reproduces_in_revision.return_value = True
    result = regression_task.validate_regression_range(
        testcase, '/a/b', 'job_type', [0, 1, 2, 3, 4], 4)
    self.assertFalse(result)

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    self.assertEqual(testcase.regression, 'NA')

  def test_valid_range(self):
    """Ensure that we handle valid ranges correctly."""
    testcase = data_types.Testcase()
    testcase.put()

    self.mock._testcase_reproduces_in_revision.return_value = False
    result = regression_task.validate_regression_range(
        testcase, '/a/b', 'job_type', [0, 1, 2, 3, 4], 4)
    self.assertTrue(result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lint changed code in current branch."""

import os
import sys

import yaml

from local.butler import appengine
from local.butler import common
from local.butler import format as formatter

_GOLINT_EXCEPTIONS = [
    'types.go'  # Not all model names conform to Go naming conventions.
]
_LICENSE_CHECK_FILENAMES = ['Dockerfile']
_LICENSE_CHECK_EXTENSIONS = [
    '.bash',
    '.c',
    '.cc',
    '.cpp',
    '.css',
    '.h',
    '.htm',
    '.html',
    '.js',
    '.go',
    '.proto',
    '.ps1',
    '.py',
    '.sh',
    '.yaml',
]
_LICENSE_CHECK_IGNORE_FILENAMES = ['technology.css']
_LICENSE_CHECK_IGNORE_DIRECTORIES = [
    'third_party',
    'templates',  # Generated code.
]
_LICENSE_CHECK_STRING = 'http://www.apache.org/licenses/LICENSE-2.0'
_LICENSE_CHECK_IGNORE = 'LICENSE_CHECK_IGNORE'
_PY_TEST_SUFFIX = '_test.py'
_PY_INIT_FILENAME = '__init__.py'
_YAML_EXCEPTIONS = ['bad.yaml']

_error_occurred = False


def _error(message=None):
  """Print error and track state via a global."""
  if message:
    print(message)

  global _error_occurred
  _error_occurred = True


def _execute_command_and_track_error(command):
  """Executes command, tracks error state."""
  returncode, output = common.execute(command, exit_on_error=False)
  if returncode != 0:
    _error()

  return output.decode('utf-8')


def license_validate(file_path):
  """Run license header validation."""
  filename = os.path.basename(file_path)
  extension = os.path.splitext(file_path)[1]
  if (filename not in _LICENSE_CHECK_FILENAMES and
      extension not in _LICENSE_CHECK_EXTENSIONS):
    return

  path_directories = file_path.split(os.sep)
  if any(d in _LICENSE_CHECK_IGNORE_DIRECTORIES for d in path_directories):
    return

  source_filename = os.path.basename(file_path)
  if source_filename in _LICENSE_CHECK_IGNORE_FILENAMES:
    return

  with open(file_path) as f:
    data = f.read()
    if _LICENSE_CHECK_STRING in data or _LICENSE_CHECK_IGNORE in data:
      return

  _error('Failed: Missing license header for %s.' % file_path)


def py_import_order(file_path):
  """Validate that python imports are alphabetized."""

  def _validate_block(import_block):
    """Ensure that a single block is ordered properly."""
    if not import_block:
      return []

    sorted_import_block = sorted(import_block, key=lambda i: i.lower())
    if sorted_import_block == import_block:
      return []

    return ['\n'.join(sorted_import_block)]

  with open(file_path) as f:
    file_content = f.read()

  imports = []
  corrected_import_blocks = []
  for line in file_content.splitlines():
    if line.startswith('import ') or line.startswith('from '):
      imports.append(line)
    else:
      corrected_import_blocks += _validate_block(imports)
      imports = []

  # Though rare, if a file ends with an import we must still validate them.
  corrected_import_blocks += _validate_block(imports)

  if not corrected_import_blocks:
    return

  suggestions = '\n\n--------\n\n'.join(corrected_import_blocks)
  _error(('Failed: File {filename} has non-alphabetized import blocks. '
          'Suggested order:\n\n{suggestions}').format(
              filename=file_path, suggestions=suggestions))


def py_test_init_check(file_path):
  """Check test directory has a __init__.py file. Otherwise, the test does not
  execute at all."""
  if not file_path.endswith(_PY_TEST_SUFFIX):
    return

  test_directory = os.path.dirname(file_path)
  if _PY_INIT_FILENAME not in os.listdir(test_directory):
    _error('Failed: Missing {filename} file in test directory {dir}.'.format(
        filename=_PY_INIT_FILENAME, dir=test_directory))


def yaml_validate(file_path):
  """Run yaml validation."""
  if os.path.basename(file_path) in _YAML_EXCEPTIONS:
    return

  try:
    with open(file_path) as f:
      yaml.safe_load(f.read())
  except Exception as e:
    _error('Failed: Invalid yaml file %s.\n\n%s' % (file_path, e))


def is_auto_generated_file(filepath):
  """Check if file is auto-generated so we dont lint it"""
  return (filepath.endswith('_pb2.py') or filepath.endswith('pb2_grpc.py') or
          os.path.dirname(filepath) == os.path.join(
              'src', 'clusterfuzz', '_internal', 'bot', 'tokenizer',
              'grammars'))


def execute(_):
  """Lint changed code."""
  pythonpath = os.getenv('PYTHONPATH', '')
  os.environ['PYTHONPATH'] = appengine.find_sdk_path() + ':' + pythonpath

  if 'GOOGLE_CLOUDBUILD' in os.environ:
    # Explicitly compare against master if we're running on the CI
    _, output = common.execute('git diff --name-only master FETCH_HEAD')
  else:
    _, output = common.execute('git diff --name-only FETCH_HEAD')

  file_paths = [
      f.decode('utf-8') for f in output.splitlines() if os.path.exists(f)
  ]
  py_changed_file_paths = [
      f for f in file_paths
      if f.endswith('.py') and not is_auto_generated_file(f)
  ]
  go_changed_file_paths = [f for f in file_paths if f.endswith('.go')]
  yaml_changed_file_paths = [f for f in file_paths if f.endswith('.yaml')]

  for file_path in py_changed_file_paths:
    line_length_override = ''
    if '_test.py' in file_path:
      line_length_override = '--max-line-length=240'

    _execute_command_and_track_error(
        f'pylint {line_length_override} {file_path}')
    _execute_command_and_track_error(f'yapf -d {file_path}')
    _execute_command_and_track_error(f'{formatter.ISORT_CMD} -c {file_path}')

    py_test_init_check(file_path)

  golint_path = os.path.join('local', 'bin', 'golint')
  for file_path in go_changed_file_paths:
    if not os.path.basename(file_path) in _GOLINT_EXCEPTIONS:
      _execute_command_and_track_error(golint_path + ' ' + file_path)

    output = _execute_command_and_track_error('gofmt -d ' + file_path)
    if output.strip():
      _error()

  for file_path in yaml_changed_file_paths:
    yaml_validate(file_path)

  for file_path in file_paths:
    license_validate(file_path)

  if _error_occurred:
    print('Linting failed, see errors above.')
    sys.exit(1)
  else:
    print('Linting passed.')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for the crash query api."""

from flask import request

from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.datastore import data_handler
from handlers import base_handler
from libs import auth
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Handler for crash querying."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  def post(self):
    """Handle a post request."""
    if not auth.get_current_user():
      raise helpers.AccessDeniedException()

    project = request.get('project')
    fuzz_target = request.get('fuzz_target')
    stacktrace = request.get('stacktrace')

    state = stack_analyzer.get_crash_data(
        stacktrace,
        symbolize_flag=False,
        fuzz_target=fuzz_target,
        already_symbolized=True,
        detect_ooms_and_hangs=True)
    security_flag = crash_analyzer.is_security_issue(
        state.crash_stacktrace, state.crash_type, state.crash_address)

    result = {
        'state': state.crash_state,
        'type': state.crash_type,
        'security': security_flag,
    }

    duplicate_testcase = data_handler.find_testcase(
        project, state.crash_type, state.crash_state, security_flag)
    if duplicate_testcase:
      result['result'] = 'duplicate'
      result['duplicate_id'] = duplicate_testcase.key.id()

      bug_id = (
          duplicate_testcase.bug_information or
          duplicate_testcase.group_bug_information)
      if bug_id:
        result['bug_id'] = str(bug_id)
    else:
      result['result'] = 'new'

    return self.render_json(result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer options."""

import configparser
import os
import random
import re

import six

from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.afl import constants as afl_constants
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

OPTIONS_FILE_EXTENSION = '.options'

# Whitelist for env variables .options files can set.
ENV_VAR_WHITELIST = set([afl_constants.DONT_DEFER_ENV_VAR])


class FuzzerOptionsException(Exception):
  """Exceptions for fuzzer options."""


class FuzzerArguments(object):
  """Fuzzer flags."""

  def __init__(self, flags):
    self.flags = flags

  def __contains__(self, key):
    return key in self.flags

  def __getitem__(self, key):
    return self.flags[key]

  def __setitem__(self, key, value):
    self.flags[key] = value

  def get(self, key, default=None, constructor=None):
    """Return value for |key|, calling the |constructor| on it, or |default| if
    the key does not exist, or if the constructor threw an exception."""
    try:
      value = self[key]
      if constructor:
        value = constructor(value)

      return value
    except Exception:
      pass

    return default

  def dict(self):
    """Return arguments as a dict."""
    return self.flags

  def list(self):
    """Return arguments as a list."""
    return ['-%s=%s' % (key, value) for key, value in six.iteritems(self.flags)]


class FuzzerOptions(object):
  """Represents fuzzer and related options."""

  OPTIONS_RANDOM_REGEX = re.compile(
      r'^\s*random\(\s*(\d+)\s*,\s*(\d+)\s*\)\s*$')

  def __init__(self, options_file_path, cwd=None):
    if not os.path.exists(options_file_path):
      raise FuzzerOptionsException('fuzzer options file does not exist.')

    if cwd:
      self._cwd = cwd
    else:
      self._cwd = os.path.dirname(options_file_path)

    self._config = configparser.ConfigParser()
    with open(options_file_path, 'r') as f:
      try:
        self._config.read_file(f)
      except configparser.Error:
        raise FuzzerOptionsException('Failed to parse fuzzer options file.')

  def _get_dict_path(self, relative_dict_path):
    """Return a full path to the dictionary."""
    return os.path.join(self._cwd, relative_dict_path)

  def _get_option_section(self, section):
    """Get an option section."""
    if not self._config.has_section(section):
      return {}

    return dict(self._config.items(section))

  def get_env(self):
    """Returns dict containing env variables and their values set by "env"
    section. Only includes env variables permitted by |ENV_VAR_WHITELIST|.
    Variables are assumed to contain no lower case letters.
    """
    env = {}
    for var_name, var_value in six.iteritems(self._get_option_section('env')):

      var_name = var_name.upper()
      if var_name in ENV_VAR_WHITELIST:
        env[var_name] = var_value

    return env

  def get_engine_arguments(self, engine):
    """Return a list of fuzzer options."""
    arguments = {}
    for option_name, option_value in six.iteritems(
        self._get_option_section(engine)):
      # Check option value for usage of random() function.
      match = self.OPTIONS_RANDOM_REGEX.match(option_value)
      if match:
        min_value, max_value = match.groups()
        option_value = str(random.SystemRandom().randint(
            int(min_value), int(max_value)))

      if option_name == 'dict':
        option_value = self._get_dict_path(option_value)

      arguments[option_name] = option_value

    return FuzzerArguments(arguments)

  def get_asan_options(self):
    """Return a list of ASAN_OPTIONS overrides."""
    return self._get_option_section('asan')

  def get_msan_options(self):
    """Return a list of MSAN_OPTIONS overrides."""
    return self._get_option_section('msan')

  def get_ubsan_options(self):
    """Return a list of UBSAN_OPTIONS overrides."""
    return self._get_option_section('ubsan')

  def get_hwasan_options(self):
    """Return a list of HWSAN_OPTIONS overrides."""
    return self._get_option_section('hwasan')

  def get_grammar_options(self):
    """Return a list og grammar options"""
    return self._get_option_section('grammar')


def get_fuzz_target_options(fuzz_target_path):
  """Return a FuzzerOptions for the given target, or None if it does not
  exist."""
  options_file_path = fuzzer_utils.get_supporting_file(fuzz_target_path,
                                                       OPTIONS_FILE_EXTENSION)

  if environment.is_trusted_host():
    options_file_path = fuzzer_utils.get_file_from_untrusted_worker(
        options_file_path)

  if not os.path.exists(options_file_path):
    return None

  options_cwd = os.path.dirname(options_file_path)

  try:
    return FuzzerOptions(options_file_path, cwd=options_cwd)
  except FuzzerOptionsException:
    logs.log_error('Invalid options file: %s.' % options_file_path)
    return None



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for path overrides"""

import os
import unittest

from clusterfuzz._internal.build_management import overrides
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'overrides_data')


def _read_data_file(filename):
  return open(os.path.join(DATA_PATH, filename)).read()


class UpdateCheckAndApplyOverridesTest(unittest.TestCase):
  """Tests for update_build_or_revision_path."""

  def setUp(self):
    test_helpers.patch(
        self, ['clusterfuzz._internal.system.environment.get_platform_id'])
    test_helpers.patch(
        self, ['clusterfuzz._internal.google_cloud_utils.storage.read_data'])
    output = _read_data_file('test_config.json')
    self.mock.read_data.return_value = output.encode()
    test_helpers.patch_environ(self)

  def test_url_branch_update(self):
    """Ensure that the path is updated correctly based on platform_id."""
    self.mock.get_platform_id.return_value = 'android:seahawk_hwasan:s'
    curr_path = 'gs://auto/config.json'
    self.assertEqual(
        overrides.check_and_apply_overrides(
            curr_path, overrides.PLATFORM_ID_TO_BUILD_PATH_KEY),
        'gs://auto/git_s/something/%something%/([something]+).zip')

  def test_empty_platform_id(self):
    """Ensure that a path is not updated if the platform_id is empty."""
    self.mock.get_platform_id.return_value = ''
    curr_path = 'gs://auto/config.json'
    self.assertRaises(overrides.BuildOverrideError,
                      overrides.check_and_apply_overrides, curr_path,
                      overrides.PLATFORM_ID_TO_BUILD_PATH_KEY)

  def test_invalid_config_url(self):
    """Ensure that a path which does not point to config.json is not updated."""
    self.mock.get_platform_id.return_value = 'android:seahawk_hwasan:r'
    curr_path = 'gs://auto/git_r/something/something/' \
                '%something%/([something]+).zip'
    self.assertEqual(
        overrides.check_and_apply_overrides(
            curr_path, overrides.PLATFORM_ID_TO_BUILD_PATH_KEY), curr_path)

  def test_unknown_platform_id(self):
    """Ensure that a path is not updated if the platform_id is unknown."""
    self.mock.get_platform_id.return_value = 'unknown'
    curr_path = 'gs://auto/config.json'
    self.assertRaises(overrides.BuildOverrideError,
                      overrides.check_and_apply_overrides, curr_path,
                      overrides.PLATFORM_ID_TO_BUILD_PATH_KEY)

  def test_empty_config(self):
    """Ensure that a path is not updated if the config is empty."""
    self.mock.get_platform_id.return_value = 'android:seahawk_hwasan:r'
    self.mock.read_data.return_value = None
    curr_path = 'gs://auto/config.json'
    self.assertRaises(overrides.BuildOverrideError,
                      overrides.check_and_apply_overrides, curr_path,
                      overrides.PLATFORM_ID_TO_BUILD_PATH_KEY)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""source_mapper tests."""
import unittest

from clusterfuzz._internal.build_management import source_mapper


class GetComponentSourceAndRelativePathTest(unittest.TestCase):
  """Tests for get_component_source_and_relative_path."""

  def test_get_component_source_and_relative_path_chromium(self):
    """Test get component source and relative path for chromium."""
    revisions_dict = {
        '/src': {
            'url': 'https://chromium.googlesource.com/chromium/src.git',
            'rev': '1d783bc2a3629b94c963debfa3feaee27092dd92',
        },
        'src/v8': {
            'url': 'https://chromium.googlesource.com/v8/v8.git',
            'rev': '7fb2c3b6db3f889ea95851ca11dcb731b07a7925',
        }
    }

    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'content/common.h', revisions_dict),
        source_mapper.ComponentPath('/src', 'content/common.h',
                                    'content/common.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'content\common.h', revisions_dict),
        source_mapper.ComponentPath('/src', 'content/common.h',
                                    'content/common.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'c:\build\src\content\common.h', revisions_dict),
        source_mapper.ComponentPath('/src', 'content/common.h',
                                    'content/common.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/mnt/build/src/content/common.h', revisions_dict),
        source_mapper.ComponentPath('/src', 'content/common.h',
                                    'content/common.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'../../third_party/WebKit/Source/platform/heap/Member.h',
            revisions_dict),
        source_mapper.ComponentPath(
            '/src', 'third_party/WebKit/Source/platform/heap/Member.h',
            'third_party/WebKit/Source/platform/heap/Member.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'v8/src/api.cc', revisions_dict),
        source_mapper.ComponentPath('src/v8', 'src/api.cc', 'v8/src/api.cc'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'v8\src\api.cc', revisions_dict),
        source_mapper.ComponentPath('src/v8', 'src/api.cc', 'v8/src/api.cc'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'c:\build\src\v8\src\api.cc', revisions_dict),
        source_mapper.ComponentPath('src/v8', 'src/api.cc', 'v8/src/api.cc'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/mnt/build/src/v8/src/api.cc', revisions_dict),
        source_mapper.ComponentPath('src/v8', 'src/api.cc', 'v8/src/api.cc'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/mnt/build/src/v8_overrides/init.cc', revisions_dict),
        source_mapper.ComponentPath('/src', 'v8_overrides/init.cc',
                                    'v8_overrides/init.cc'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/mnt/build/non_existent', revisions_dict),
        source_mapper.ComponentPath())
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'c:\build\non_existent', revisions_dict),
        source_mapper.ComponentPath())

  def test_get_component_source_and_relative_path_oss_fuzz(self):
    """Test get comnponent source and relative path for OSS-Fuzz."""
    revisions_dict = {
        '/src/libass': {
            'url': 'https://github.com/libass/libass.git',
            'rev': '35dc4dd0e14e3afb4a2c7e319a3f4110e20c7cf2',
            'type': 'git'
        },
        '/src/fribidi': {
            'url': 'https://github.com/behdad/fribidi.git',
            'rev': '881b8d891cc61989ab8811b74d0e721f72bf913b',
            'type': 'git'
        }
    }

    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/src/libass/test/test.c', revisions_dict),
        source_mapper.ComponentPath('/src/libass', 'test/test.c',
                                    'libass/test/test.c'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/src/fribidi/lib/common.h', revisions_dict),
        source_mapper.ComponentPath('/src/fribidi', 'lib/common.h',
                                    'fribidi/lib/common.h'))
    self.assertEqual(
        source_mapper.get_component_source_and_relative_path(
            r'/src/not_existent', revisions_dict),
        source_mapper.ComponentPath())


class NormalizeSourcePathTest(unittest.TestCase):
  """Tests for normalize_source_path."""

  def test_normalize_source_path(self):
    """Test normalizing source path."""
    self.assertEqual(
        source_mapper.normalize_source_path(r'c:\build\src\heap\compact.cc'),
        'heap/compact.cc')
    self.assertEqual(
        source_mapper.normalize_source_path('/mnt/build/src/heap/compact.cc'),
        'heap/compact.cc')
    self.assertEqual(
        source_mapper.normalize_source_path('/mnt/build/heap/compact.cc'), None)
    self.assertEqual(
        source_mapper.normalize_source_path('src/heap/compact.cc'),
        'src/heap/compact.cc')
    self.assertEqual(
        source_mapper.normalize_source_path('heap/compact.cc'),
        'heap/compact.cc')
    self.assertEqual(
        source_mapper.normalize_source_path('/proc/self/cwd/heap/compact.cc'),
        'heap/compact.cc')
    self.assertEqual(
        source_mapper.normalize_source_path(
            '../../third_party/WebKit/Source/platform/heap/Member.h'),
        'third_party/WebKit/Source/platform/heap/Member.h')


class VCSViewerTest(unittest.TestCase):
  """Tests for VCSViewer."""

  def test_get_vcs_viewer_for_url(self):
    """Test that VCS recognition logic from source_mapper.py works correctly."""
    vcs = source_mapper.get_vcs_viewer_for_url(
        'https://anongit.freedesktop.org/git/gstreamer/gstreamer.git')
    self.assertIsInstance(vcs, source_mapper.FreeDesktopVCS)

    vcs = source_mapper.get_vcs_viewer_for_url(
        'https://github.com/imagemagick/imagemagick.git')
    self.assertIsInstance(vcs, source_mapper.GitHubVCS)

    vcs = source_mapper.get_vcs_viewer_for_url(
        'https://gitlab.com/libidn/libidn2.git')
    self.assertIsInstance(vcs, source_mapper.GitLabVCS)

    vcs = source_mapper.get_vcs_viewer_for_url(
        'https://gitlab.gnome.org/GNOME/libxml2.git')
    self.assertIsInstance(vcs, source_mapper.GitLabVCS)

    vcs = source_mapper.get_vcs_viewer_for_url(
        'https://chromium.googlesource.com/webm/libwebp.git')
    self.assertIsInstance(vcs, source_mapper.GoogleSourceVCS)

    vcs = source_mapper.get_vcs_viewer_for_url('//dir1/dir2')
    self.assertIsInstance(vcs, source_mapper.GoogleVCS)

    vcs = source_mapper.get_vcs_viewer_for_url(
        'http://hg.code.sf.net/p/graphicsmagick/code')
    self.assertIsInstance(vcs, source_mapper.MercurialVCS)

  def test_google_vcs(self):
    vcs = source_mapper.GoogleVCS('//dir1/dir2')
    self.assertEqual('https://cs.corp.google.com/dir1/dir2/?rcl=1337',
                     vcs.get_source_url_for_revision(1337))
    self.assertEqual(None, vcs.get_source_url_for_revision_diff(1337, 1338))
    self.assertEqual(
        'https://cs.corp.google.com'
        '/dir1/dir2/dir3/dir4/file1.txt?rcl=1337&l=10',
        vcs.get_source_url_for_revision_path_and_line(
            1337, 'dir3/dir4/file1.txt', 10))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""variant_task tests."""
import unittest

from clusterfuzz._internal.bot.tasks import variant_task
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class GetVariantTestcaseForJob(unittest.TestCase):
  """Test _get_variant_testcase_for_job."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_same_job(self):
    """Test variant task with same job."""
    testcase = test_utils.create_generic_testcase()
    variant_testcase = variant_task._get_variant_testcase_for_job(  # pylint: disable=protected-access
        testcase, testcase.job_type)
    self.assertEqual(testcase, variant_testcase)

  def test_blackbox_fuzzer_job(self):
    """Test variant task with blackbox fuzzer job."""
    testcase = test_utils.create_generic_testcase()
    testcase.job_type = 'linux_asan_project'
    testcase.put()

    variant_testcase = variant_task._get_variant_testcase_for_job(  # pylint: disable=protected-access
        testcase, 'linux_msan_project')
    self.assertEqual(testcase, variant_testcase)

  def test_engine_fuzzer_job(self):
    """Test variant task with an engine fuzzer job."""
    testcase = data_types.Testcase(
        job_type='libfuzzer_asan_project',
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libfuzzer_project_binary_name',
        project_name='project',
        crash_type='crash-type',
        crash_address='0x1337',
        crash_state='A\nB\nC\n',
        crash_revision=1337)
    testcase.set_metadata(
        'fuzzer_binary_name', 'binary_name', update_testcase=True)

    job = data_types.Job()
    job.name = 'afl_asan_project'
    job.environment_string = 'PROJECT_NAME = project\n'
    job.put()

    variant_testcase = variant_task._get_variant_testcase_for_job(  # pylint: disable=protected-access
        testcase, 'afl_asan_project')
    self.assertNotEqual(testcase, variant_testcase)
    self.assertEqual(testcase.key.id(), variant_testcase.key.id())
    self.assertEqual('afl', variant_testcase.fuzzer_name)
    self.assertEqual('afl_project_binary_name',
                     variant_testcase.overridden_fuzzer_name)
    self.assertEqual('afl_asan_project', variant_testcase.job_type)

    self.assertEqual('crash-type', variant_testcase.crash_type)
    self.assertEqual('0x1337', variant_testcase.crash_address)
    self.assertEqual('A\nB\nC\n', variant_testcase.crash_state)
    self.assertEqual(1337, variant_testcase.crash_revision)
    self.assertEqual('binary_name',
                     variant_testcase.get_metadata('fuzzer_binary_name'))

    # Test that a put() call does not change original testcase.
    variant_testcase.comments = 'ABC'
    variant_testcase.put()
    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    self.assertEqual('', testcase.comments)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for creating issue."""

from flask import request

from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers
from libs.issue_management import issue_filer


class Handler(base_handler.Handler):
  """Handler that creates an issue."""

  @staticmethod
  def create_issue(testcase, severity, cc_me):
    """Create an issue."""
    issue_tracker = helpers.get_issue_tracker_for_testcase(testcase)
    user_email = helpers.get_user_email()

    if severity is not None:
      severity = helpers.cast(
          severity, int, 'Invalid value for security severity (%s).' % severity)

    additional_ccs = []
    if cc_me:
      additional_ccs.append(user_email)

    issue_id = issue_filer.file_issue(
        testcase,
        issue_tracker,
        security_severity=severity,
        user_email=user_email,
        additional_ccs=additional_ccs)

    if not issue_id:
      raise helpers.EarlyExitException('Unable to create new issue.', 500)

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_testcase_access
  def post(self, testcase):
    """Create an issue."""
    cc_me = request.get('ccMe')
    severity = request.get('severity')

    self.create_issue(testcase, severity, cc_me)
    return self.render_json(show.get_testcase_detail(testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Train ml rnn model."""

import argparse
import math
import os
import sys
import time

import numpy as np
import tensorflow as tf

from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.bot.fuzzers.ml.rnn import utils

# Training suggestions
#
# Training only:
#     Leave all the parameters as they are in constants.py.
#     Disable validation to run a bit faster (set validation=False).
#     You can follow progress in Tensorboard: tensorboard --logdir=log
#
# Training and experimenting (default):
#     Keep validation enabled.
#     You can now play with the parameters and follow the effects in
#     Tensorboard.
#     A good choice of parameters ensures that the testing and validation
#     curves stay close. To see the curves drift apart ("overfitting") try
#     to use an insufficient amount of training data.


@tf.function
def train_step(model, optimizer, input_data, expected_data, train=False):
  """Train the model for one step.

  Args:
    model: RNN model to train/predict.
    optimize: optimizer to use to train the model.
    input_data: input sequence to the model.
    expected_data: expected output of the model.

  Returns:
    Tuple containing the sequential loss between the expected output and the
    real output, the batch loss between the two, the accuracy metric value as
    well as the most likely predicted output.
  """
  with tf.GradientTape() as tape:
    predicted_data = model(input_data)
    loss = tf.keras.losses.sparse_categorical_crossentropy(
        expected_data, predicted_data, from_logits=True)
    seq_loss = tf.reduce_mean(input_tensor=loss, axis=1)
    batch_loss = tf.reduce_mean(input_tensor=seq_loss)

    output_bytes = tf.cast(
        tf.argmax(predicted_data, axis=-1), expected_data.dtype)
    accuracy = tf.reduce_mean(
        tf.cast(tf.equal(expected_data, output_bytes), tf.float32))

  if train:
    grads = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(grads, model.trainable_variables))

  return seq_loss, batch_loss, accuracy, output_bytes


def main(args):
  """Main function to train the model.

  Args:
    args: Parsed arguments.

  Returns:
    Execution status defined by `constants.ExitCode`.
  """
  # Validate paths.
  if not validate_paths(args):
    return constants.ExitCode.INVALID_PATH

  # Extract paths.
  input_dir = args.input_dir
  model_dir = args.model_dir
  log_dir = args.log_dir
  existing_model = args.existing_model

  # Extract model parameters.
  batch_size = args.batch_size
  dropout_pkeep = args.dropout_pkeep
  hidden_state_size = args.hidden_state_size
  hidden_layer_size = args.hidden_layer_size
  learning_rate = args.learning_rate

  # Extract additional flags.
  debug = args.debug
  validation = args.validation

  # Split corpus for training and validation.
  # validation_text will be empty if validation is False.
  code_text, validation_text, input_ranges = utils.read_data_files(
      input_dir, validation=validation)

  # Bail out if we don't have enough corpus for training.
  if len(code_text) < batch_size * constants.TRAINING_SEQLEN + 1:
    return constants.ExitCode.CORPUS_TOO_SMALL

  # Get corpus files info. Will be used in debug mode to generate sample text.
  files_info_list = []
  if debug:
    files_info_list = utils.get_files_info(input_dir)
    assert files_info_list

  # Calculate validation batch size. It will be 0 if we choose not to validate.
  validation_batch_size = len(validation_text) // constants.VALIDATION_SEQLEN

  # Display some stats on the data.
  epoch_size = len(code_text) // (batch_size * constants.TRAINING_SEQLEN)
  utils.print_data_stats(len(code_text), len(validation_text), epoch_size)

  # Set global random seed, so any random sequence generated is repeatable.
  # It could also be removed.
  tf.random.set_seed(0)

  # Build the RNN model.
  model = utils.build_model(hidden_layer_size * hidden_state_size,
                            dropout_pkeep, batch_size, debug)

  # Choose Adam optimizer to compute gradients.
  optimizer = tf.keras.optimizers.Adam(learning_rate)

  # Init Tensorboard stuff.
  # This will save Tensorboard information in folder specified in command line.
  # Two sets of data are saved so that you can compare training and
  # validation curves visually in Tensorboard.
  timestamp = str(math.trunc(time.time()))
  summary_writer = tf.summary.create_file_writer(
      os.path.join(log_dir, timestamp + '-training'))
  validation_writer = tf.summary.create_file_writer(
      os.path.join(log_dir, timestamp + '-validation'))

  # For display: init the progress bar.
  step_size = batch_size * constants.TRAINING_SEQLEN
  frequency = constants.DISPLAY_FREQ * step_size
  progress = utils.Progress(
      constants.DISPLAY_FREQ,
      size=constants.DISPLAY_LEN,
      msg='Training on next {} batches'.format(constants.DISPLAY_FREQ))

  # We continue training on existing model, or start with a new model.
  if existing_model:
    print('Continue training on existing model: {}'.format(existing_model))
    try:
      model.load_weights(existing_model)
    except:
      print(
          ('Failed to restore existing model since model '
           'parameters do not match.'),
          file=sys.stderr)
      return constants.ExitCode.TENSORFLOW_ERROR
  else:
    print('No existing model provided. Start training with a new model.')

  # Num of bytes we have trained so far.
  steps = 0

  # Training loop.
  for input_batch, expected_batch, epoch in utils.rnn_minibatch_sequencer(
      code_text,
      batch_size,
      constants.TRAINING_SEQLEN,
      nb_epochs=constants.EPOCHS):

    # Train on one mini-batch.
    seq_loss, batch_loss, accuracy, output_bytes = train_step(
        model, optimizer, input_batch, expected_batch, train=True)

    # Log training data for Tensorboard display a mini-batch of sequences
    # every `frequency` batches.
    if debug and steps % frequency == 0:
      utils.print_learning_learned_comparison(
          input_batch, output_bytes, seq_loss, input_ranges, batch_loss,
          accuracy, epoch_size, steps, epoch)
      with summary_writer.as_default():  # pylint: disable=not-context-manager
        tf.summary.scalar('batch_loss', batch_loss, step=steps)
        tf.summary.scalar('batch_accuracy', accuracy, step=steps)
      summary_writer.flush()

    # Run a validation step every `frequency` batches.
    # The validation text should be a single sequence but that's too slow.
    # We cut it up and batch the pieces (slightly inaccurate).
    if validation and steps % frequency == 0 and validation_batch_size:
      utils.print_validation_header(len(code_text), input_ranges)
      validation_x, validation_y, _ = next(
          utils.rnn_minibatch_sequencer(validation_text, validation_batch_size,
                                        constants.VALIDATION_SEQLEN, 1))

      validation_model = utils.build_model(
          hidden_layer_size * hidden_state_size, dropout_pkeep,
          validation_batch_size, False)
      last_weights = tf.train.latest_checkpoint(model_dir)
      if last_weights:
        validation_model.load_weights(tf.train.latest_checkpoint(model_dir))
        validation_model.build(tf.TensorShape([validation_batch_size, None]))
        validation_model.reset_states()

      # Run one single inference step
      _, batch_loss, accuracy, _ = train_step(
          validation_model, optimizer, validation_x, validation_y, train=False)

      utils.print_validation_stats(batch_loss, accuracy)

      # Save validation data for Tensorboard.
      with validation_writer.as_default():  # pylint: disable=not-context-manager
        tf.summary.scalar('batch_loss', batch_loss, step=steps)
        tf.summary.scalar('batch_accuracy', accuracy, step=steps)
      validation_writer.flush()

    # Display a short text generated with the current weights and biases.
    # If enabled, there will be a large output.
    if debug and steps // 4 % frequency == 0:
      utils.print_text_generation_header()
      file_info = utils.random_element_from_list(files_info_list)
      first_byte, file_size = file_info['first_byte'], file_info['file_size']
      ry = np.array([[first_byte]])
      sample = [first_byte]

      generation_model = utils.build_model(
          hidden_layer_size * hidden_state_size, dropout_pkeep, 1, False)
      last_weights = tf.train.latest_checkpoint(model_dir)
      if last_weights:
        generation_model.load_weights(tf.train.latest_checkpoint(model_dir))
        generation_model.build(tf.TensorShape([1, None]))
        generation_model.reset_states()

      for _ in range(file_size - 1):
        prediction = generation_model(ry)
        prediction = tf.squeeze(prediction, 0).numpy()
        rc = utils.sample_from_probabilities(
            prediction, topn=10 if epoch <= 1 else 2)
        sample.append(rc)
        ry = np.array([[rc]])

      print(repr(utils.decode_to_text(sample)))
      utils.print_text_generation_footer()

    # Save a checkpoint every `10 * frequency` batches. Each checkpoint is
    # a version of model.
    if steps // 10 % frequency == 0:
      saved_model_name = constants.RNN_MODEL_NAME + '_' + timestamp
      saved_model_path = os.path.join(model_dir, saved_model_name)
      model.save_weights(saved_model_path)
      print('Saved model: {}'.format(saved_model_path))

    # Display progress bar.
    if debug:
      progress.step(reset=steps % frequency == 0)

    # Update state.
    steps += step_size

  # Save the model after training is done.
  saved_model_name = constants.RNN_MODEL_NAME + '_' + timestamp
  saved_model_path = os.path.join(model_dir, saved_model_name)
  model.save_weights(saved_model_path)
  print('Saved model: {}'.format(saved_model_path))

  return constants.ExitCode.SUCCESS


def validate_paths(args):
  """Validate paths.

  Args:
    args: Parsed arguments.

  Returns:
    True if all paths are valid, False otherwise.
  """
  if not os.path.exists(args.input_dir):
    print(
        'Input directory {} does not exist'.format(args.input_dir),
        file=sys.stderr)
    return False

  if not os.path.exists(args.model_dir):
    os.mkdir(args.model_dir)

  if not os.path.exists(args.log_dir):
    os.mkdir(args.log_dir)

  if args.existing_model and not utils.validate_model_path(args.existing_model):
    print(
        'Existing model {} does not exist'.format(args.existing_model),
        file=sys.stderr)
    return False

  return True


def parse_args():
  """Parse command line arguments.

  Returns:
    Parsed arguement object.
  """
  parser = argparse.ArgumentParser('Training RNN model on existing testcases')

  parser.add_argument('--input-dir', help='Input folder path', required=True)
  parser.add_argument('--log-dir', help='Log folder path', required=True)
  parser.add_argument('--model-dir', help='Path to save models', required=True)

  # Optional arguments: model parameters and additional flags.
  parser.add_argument(
      '--batch-size', help='Batch size', type=int, default=constants.BATCH_SIZE)
  parser.add_argument(
      '--debug', help='Print training progress', action='store_true')
  parser.add_argument(
      '--dropout-pkeep',
      help='Dropout probability (keep rate)',
      type=float,
      default=constants.DROPOUT_PKEEP)
  parser.add_argument(
      '--existing-model', help='Continue training on existing model')
  parser.add_argument(
      '--hidden-state-size',
      help='Hidden state size of LSTM cell',
      type=int,
      default=constants.HIDDEN_STATE_SIZE)
  parser.add_argument(
      '--hidden-layer-size',
      help='Hidden layer size of LSTM model',
      type=int,
      default=constants.HIDDEN_LAYER_SIZE)
  parser.add_argument(
      '--learning-rate',
      help='Learning rate',
      type=float,
      default=constants.LEARNING_RATE)
  parser.add_argument(
      '--validation',
      help='Print validation stats during training',
      action='store_true')

  return parser.parse_args()


if __name__ == '__main__':
  parsed_args = parse_args()
  sys.exit(main(parsed_args))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for external_users."""

import datetime
import unittest

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class ExternalUsersTest(unittest.TestCase):
  """External users test."""

  def setUp(self):
    helpers.patch_environ(self)

    # Fake permissions.
    data_types.ExternalUserPermission(
        email='user@example.com',
        entity_name='fuzzer',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user2@example.com',
        entity_name='fuzz',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.SECURITY).put()

    data_types.ExternalUserPermission(
        email='user3@example.com',
        entity_name='parent_',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user4@example.com',
        entity_name='parent',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user5@example.com',
        entity_name='parent_cg',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user6@example.com',
        entity_name='parens',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user7@example.com',
        entity_name='parent',
        entity_kind=data_types.PermissionEntityKind.FUZZER,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user2@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=True,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user3@example.com',
        entity_name='job2',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user8@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user9@example.com',
        entity_name='job2',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user10@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='user10@example.com',
        entity_name='job3',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.NONE).put()

    data_types.ExternalUserPermission(
        email='uploader1@example.com',
        entity_name=None,
        entity_kind=data_types.PermissionEntityKind.UPLOADER,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.NONE).put()

    # Fake fuzzers.
    data_types.Fuzzer(name='fuzzer').put()
    data_types.Fuzzer(name='parent', jobs=['job', 'job2', 'job3']).put()

    data_types.Job(name='job').put()
    data_types.Job(name='job2').put()
    data_types.Job(name='job3').put()

    data_types.FuzzTarget(
        engine='parent', binary='child', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='parent_child',
        job='job',
        last_run=datetime.datetime.utcnow()).put()

    data_types.FuzzTarget(
        engine='parent', binary='child2', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='parent_child2',
        job='job',
        last_run=datetime.datetime.utcnow()).put()

    data_types.FuzzTarget(
        engine='parent', binary='child', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='parent_child',
        job='job3',
        last_run=datetime.datetime.utcnow()).put()

  def test_allowed_fuzzers(self):
    """allowed_fuzzers_for_user tests."""
    # Direct match.
    result = external_users.allowed_fuzzers_for_user('User@example.com')
    self.assertEqual(result, ['fuzzer'])

    # Prefix on fuzzer name.
    result = external_users.allowed_fuzzers_for_user('uSer2@example.com')
    self.assertEqual(result, ['fuzzer'])

    # Prefix on child fuzzer name.
    result = external_users.allowed_fuzzers_for_user('user3@example.com')
    self.assertEqual(result, ['parent_child', 'parent_child2'])

    # Direct match on a parent fuzzer that has children. Should not have any
    # results.
    result = external_users.allowed_fuzzers_for_user('user4@example.com')
    self.assertEqual(len(result), 0)

    # No such user.
    result = external_users.allowed_fuzzers_for_user('notexist@example.com')
    self.assertEqual(result, [])

  def test_allowed_users_for_fuzzer(self):
    """allowed_users_for_fuzzer tests."""
    # Direct match + a prefix match.
    result = external_users.allowed_users_for_fuzzer('fuzzer')
    self.assertEqual(result, ['user2@example.com', 'user@example.com'])

    # Child fuzzer prefix match.
    result = external_users.allowed_users_for_fuzzer('parent_child')
    self.assertEqual(result, ['user3@example.com', 'user7@example.com'])

  def test_is_fuzzer_allowed_for_user(self):
    """is_fuzzer_allowed_for_user tests."""
    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user('uSer@example.com', 'fuzzer'))
    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user('useR2@example.com',
                                                  'fuzzer'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user3@example.com',
                                                  'fuzzer'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user4@example.com',
                                                  'fuzzer'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user5@example.com',
                                                  'fuzzer'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user6@example.com',
                                                  'fuzzer'))
    # No such user.
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('notexist@example.com',
                                                  'fuzzer'))

    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user1@example.com',
                                                  'parent_child'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user2@example.com',
                                                  'parent_child'))
    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user('user3@example.com',
                                                  'parent_child'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user4@example.com',
                                                  'parent_child'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user5@example.com',
                                                  'parent_child'))
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('user6@example.com',
                                                  'parent_child'))
    # No such user.
    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user('notexist@example.com',
                                                  'fuzzer'))

  def test_is_fuzzer_allowed_for_user_including_jobs(self):
    """is_fuzzer_allowed_for_user tests with include_from_jobs == True."""
    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user(
            'User@example.com', 'parent_child', include_from_jobs=True))

    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user(
            'uSer@example.com', 'parent_child', include_from_jobs=False))

    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user(
            'uSer@example.com', 'parent_child2', include_from_jobs=True))

    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user(
            'user@example.com', 'parent_child2', include_from_jobs=False))

    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user(
            'user@example.com', 'parent_child3', include_from_jobs=True))

    self.assertFalse(
        external_users.is_fuzzer_allowed_for_user(
            'user@example.com', 'parent', include_from_jobs=True))

    self.assertTrue(
        external_users.is_fuzzer_allowed_for_user(
            'user@example.com', 'fuzzer', include_from_jobs=True))

  def test_is_job_allowed_for_user(self):
    """is_job_allowed_for_user tests."""
    self.assertTrue(
        external_users.is_job_allowed_for_user('useR@example.com', 'job'))
    self.assertTrue(
        external_users.is_job_allowed_for_user('user2@example.com', 'job'))
    self.assertTrue(
        external_users.is_job_allowed_for_user('user8@example.com', 'job'))
    self.assertFalse(
        external_users.is_job_allowed_for_user('user3@example.com', 'job'))
    self.assertTrue(
        external_users.is_job_allowed_for_user('User3@example.com', 'job2'))
    self.assertTrue(
        external_users.is_job_allowed_for_user('user9@example.com', 'job2'))
    self.assertFalse(
        external_users.is_job_allowed_for_user('user@example.com', 'job2'))

  def test_is_upload_allowed_for_user(self):
    """is_upload_allowed_for_user tests."""
    self.assertTrue(
        external_users.is_upload_allowed_for_user('uploader1@example.com'))
    self.assertFalse(
        external_users.is_upload_allowed_for_user('user@example.com'))

  def test_cc_users_for_fuzzer(self):
    """cc_users_for_fuzzer tests."""
    result = external_users.cc_users_for_fuzzer('fuzzer', security_flag=False)
    self.assertEqual(result, ['user@example.com'])

    result = external_users.cc_users_for_fuzzer('fuzzer', security_flag=True)
    self.assertEqual(result, ['user2@example.com', 'user@example.com'])

    result = external_users.cc_users_for_fuzzer(
        'parent_child', security_flag=True)
    self.assertListEqual(result, ['user7@example.com'])

  def test_cc_users_for_job(self):
    """cc_users_for_job tests."""
    result = external_users.cc_users_for_job('job', security_flag=False)
    self.assertEqual(result, ['user2@example.com', 'user@example.com'])

    result = external_users.cc_users_for_job('job', security_flag=True)
    self.assertEqual(result, ['user2@example.com', 'user@example.com'])

    result = external_users.cc_users_for_job('job2', security_flag=False)
    self.assertEqual(result, ['user2@example.com', 'user3@example.com'])

  def test_allowed_jobs(self):
    """Test allowed_jobs_for_user."""
    result = external_users.allowed_jobs_for_user('User@example.com')
    self.assertEqual(['job'], result)

    result = external_users.allowed_jobs_for_user('uSer2@example.com')
    self.assertEqual(['job', 'job2', 'job3'], result)

    result = external_users.allowed_jobs_for_user('user3@example.com')
    self.assertEqual(['job2'], result)

    result = external_users.allowed_jobs_for_user('user4@example.com')
    self.assertEqual([], result)

  def test_allowed_fuzzers_including_jobs(self):
    """Tests allowed_fuzzers_for_user with jobs."""
    result = external_users.allowed_fuzzers_for_user(
        'User@example.com', include_from_jobs=True)
    self.assertEqual(['fuzzer', 'parent_child', 'parent_child2'], result)
    result = external_users.allowed_fuzzers_for_user(
        'User8@example.com', include_from_jobs=True)
    self.assertEqual(['parent_child', 'parent_child2'], result)
    result = external_users.allowed_fuzzers_for_user(
        'user8@example.com', include_from_jobs=True, include_parents=True)
    self.assertEqual(['parent', 'parent_child', 'parent_child2'], result)
    result = external_users.allowed_fuzzers_for_user(
        'user9@example.com', include_from_jobs=True)
    self.assertEqual([], result)

    result = external_users.allowed_fuzzers_for_user('user10@example.com')
    self.assertEqual([], result)

    result = external_users.allowed_fuzzers_for_user(
        'user10@example.com', include_from_jobs=True)
    self.assertEqual(['parent_child', 'parent_child2'], result)

    result = external_users.allowed_fuzzers_for_user(
        'user10@example.com', include_from_jobs=True, include_parents=True)
    self.assertEqual(['parent', 'parent_child', 'parent_child2'], result)


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Impact task.
   Determine whether or not a test case affects production branches."""

import six

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.chrome import build_info
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment


class BuildFailedException(Exception):
  """Represents the build failure."""


class AppFailedException(Exception):
  """Represents the app failure."""


class Impact(object):
  """Represents impact on a build type."""

  def __init__(self, version='', likely=False, extra_trace=''):
    self.version = str(version)
    self.likely = likely
    self.extra_trace = extra_trace

  def is_empty(self):
    """Return True if empty."""
    return not self.version

  def __eq__(self, other):
    return (self.version == other.version and self.likely == other.likely and
            self.extra_trace == other.extra_trace)


class Impacts(object):
  """Represents impacts on different release channels."""

  def __init__(self, stable=None, beta=None, extended_stable=None, head=None):
    self.stable = stable or Impact()
    self.beta = beta or Impact()
    self.extended_stable = extended_stable or Impact()
    self.head = head or Impact()

  def is_empty(self):
    return (self.extended_stable.is_empty() and self.stable.is_empty() and
            self.beta.is_empty() and self.head.is_empty())

  def get_extra_trace(self):
    return (
        self.extended_stable.extra_trace + '\n' + self.stable.extra_trace + '\n'
        + self.beta.extra_trace + '\n' + self.head.extra_trace).strip()

  def __eq__(self, other):
    return (self.extended_stable == other.extended_stable and
            self.stable == other.stable and self.beta == other.beta and
            self.head == other.head)


def get_chromium_component_start_and_end_revision(start_revision, end_revision,
                                                  job_type):
  """Get revisions from chromium component."""
  component_rev_list = revisions.get_component_range_list(
      start_revision, end_revision, job_type)

  for component_rev in component_rev_list:
    if component_rev['component'] == 'Chromium':
      start_revision, end_revision = (
          revisions.get_start_and_end_revision(component_rev['link_text']))

  return start_revision, end_revision


def get_start_and_end_revision(regression_range, job_type):
  """Get start and end revision."""
  start_revision, end_revision = revisions.get_start_and_end_revision(
      regression_range)

  # FIXME: Hack to use chromium revision for android builds.
  if environment.is_android():
    return get_chromium_component_start_and_end_revision(
        start_revision, end_revision, job_type)

  return start_revision, end_revision


def get_component_information_by_name(chromium_revision,
                                      component_display_name):
  """Returns a dictionary with information about a component at a revision."""
  lower_name = component_display_name.lower()
  component_revisions = revisions.get_component_revisions_dict(
      chromium_revision, None)
  if component_revisions is None:
    return None

  all_details = []
  for value in six.itervalues(component_revisions):
    if value and 'name' in value and value['name'].lower() == lower_name:
      all_details.append(value)
  # If we found several components with the same name, return nothing useful.
  if len(all_details) == 1:
    return all_details[0]
  return None


def get_component_impacts_from_url(component_name,
                                   regression_range,
                                   job_type,
                                   platform=None):
  """Gets component impact string using the build information url."""
  start_revision, end_revision = get_start_and_end_revision(
      regression_range, job_type)
  if not end_revision:
    return Impacts()

  build_revision_mappings = build_info.get_build_to_revision_mappings(platform)
  if not build_revision_mappings:
    return Impacts()

  found_impacts = dict()
  for build in ['extended_stable', 'stable', 'beta', 'canary']:
    mapping = build_revision_mappings.get(build)
    # TODO(yuanjunh): bypass for now but remove it after ES is enabled.
    if build == 'extended_stable' and not mapping:
      found_impacts[build] = Impact()
      continue
    # Some platforms don't have canary, so use dev to represent
    # the affected head version.
    if build == 'canary' and not mapping:
      mapping = build_revision_mappings.get('dev')
    if not mapping:
      return Impacts()
    chromium_revision = mapping['revision']
    component_revision = get_component_information_by_name(
        chromium_revision, component_name)
    if not component_revision:
      return Impacts()
    branched_from = revisions.revision_to_branched_from(
        component_revision['url'], component_revision['rev'])
    if not branched_from:
      return Impacts()
    impact = get_impact({
        'revision': branched_from,
        'version': mapping['version']
    }, start_revision, end_revision, build == 'canary')
    found_impacts[build] = impact
  return Impacts(found_impacts['stable'], found_impacts['beta'],
                 found_impacts['extended_stable'], found_impacts['canary'])


def get_impacts_from_url(regression_range, job_type, platform=None):
  """Gets impact string using the build information url."""
  component_name = data_handler.get_component_name(job_type)
  if component_name:
    return get_component_impacts_from_url(component_name, regression_range,
                                          job_type, platform)

  start_revision, end_revision = get_start_and_end_revision(
      regression_range, job_type)
  if not end_revision:
    return Impacts()

  build_revision_mappings = build_info.get_build_to_revision_mappings(platform)
  if not build_revision_mappings:
    return Impacts()

  extended_stable = get_impact(
      build_revision_mappings.get('extended_stable'), start_revision,
      end_revision)
  stable = get_impact(
      build_revision_mappings.get('stable'), start_revision, end_revision)
  beta = get_impact(
      build_revision_mappings.get('beta'), start_revision, end_revision)
  head = get_head_impact(build_revision_mappings, start_revision, end_revision)

  return Impacts(stable, beta, extended_stable, head)


def get_impact(build_revision,
               start_revision,
               end_revision,
               is_last_possible_build=False):
  """Return a Impact object represents the impact on a given build_type. Or
    return None."""
  if not build_revision:
    return Impact()

  revision = build_revision['revision']
  if not revision.isdigit():
    return Impact()

  revision = int(revision)

  version = build_revision['version']
  if start_revision > revision:
    if is_last_possible_build:
      # There are no further builds to be tested. We are probably testing
      # a revision of the code which hasn't yet made it into *any* build.
      # If that's the case, we'll say that this test case _probably_
      # impacts the milestone. We can't be sure, because the next build
      # might happen to gain a new milestone number, but it's unlikely.
      milestone = version.split('.')[0]
      return Impact(milestone, likely=True)
    return Impact()

  if end_revision < revision:
    return Impact(version, likely=False)

  # We can't figure out the impact, but it is likely.
  return Impact(version, likely=True)


def get_impacts_on_prod_builds(testcase, testcase_file_path):
  """Get testcase impact on production builds, which are extended stable, stable
  and beta."""
  impacts = Impacts()
  try:
    impacts.stable = get_impact_on_build(
        'stable', testcase.impact_stable_version, testcase, testcase_file_path)
  except AppFailedException:
    return get_impacts_from_url(testcase.regression, testcase.job_type)

  try:
    impacts.beta = get_impact_on_build('beta', testcase.impact_beta_version,
                                       testcase, testcase_file_path)
  except AppFailedException:
    # If beta fails to get the binary, we ignore. At least, we have stable.
    pass

  try:
    impacts.extended_stable = get_impact_on_build(
        'extended_stable', testcase.impact_extended_stable_version, testcase,
        testcase_file_path)
  except AppFailedException:
    return get_impacts_from_url(testcase.regression, testcase.job_type)

  # Always record the affected head version.
  start_revision, end_revision = get_start_and_end_revision(
      testcase.regression, testcase.job_type)
  build_revision_mappings = build_info.get_build_to_revision_mappings()
  impacts.head = get_head_impact(build_revision_mappings, start_revision,
                                 end_revision)

  return impacts


def get_head_impact(build_revision_mappings, start_revision, end_revision):
  """Return the impact on 'head', i.e. the latest build we can find."""
  latest_build = build_revision_mappings.get('canary')
  if latest_build is None:
    latest_build = build_revision_mappings.get('dev')
  return get_impact(
      latest_build, start_revision, end_revision, is_last_possible_build=True)


def get_impact_on_build(build_type, current_version, testcase,
                        testcase_file_path):
  """Return impact and additional trace on a prod build given build_type."""
  build = build_manager.setup_production_build(build_type)
  if not build:
    raise BuildFailedException(
        'Build setup failed for %s' % build_type.capitalize())

  if not build_manager.check_app_path():
    raise AppFailedException()

  version = build.revision
  if version == current_version:
    return Impact(current_version, likely=False)

  app_path = environment.get_value('APP_PATH')
  command = testcase_manager.get_command_line_for_application(
      testcase_file_path, app_path=app_path, needs_http=testcase.http_flag)

  result = testcase_manager.test_for_crash_with_retries(
      testcase,
      testcase_file_path,
      environment.get_value('TEST_TIMEOUT'),
      http_flag=testcase.http_flag)

  if result.is_crash():
    symbolized_crash_stacktrace = result.get_stacktrace(symbolized=True)
    unsymbolized_crash_stacktrace = result.get_stacktrace(symbolized=False)
    stacktrace = utils.get_crash_stacktrace_output(
        command, symbolized_crash_stacktrace, unsymbolized_crash_stacktrace,
        build_type)
    return Impact(version, likely=False, extra_trace=stacktrace)

  return Impact()


def set_testcase_with_impacts(testcase, impacts):
  """Set testcase's impact-related fields given impacts."""
  testcase.impact_extended_stable_version = impacts.extended_stable.version
  testcase.impact_extended_stable_version_likely = \
    impacts.extended_stable.likely
  testcase.impact_stable_version = impacts.stable.version
  testcase.impact_stable_version_likely = impacts.stable.likely
  testcase.impact_beta_version = impacts.beta.version
  testcase.impact_beta_version_likely = impacts.beta.likely
  testcase.impact_head_version = impacts.head.version
  testcase.impact_head_version_likely = impacts.head.likely
  testcase.is_impact_set_flag = True


def execute_task(testcase_id, job_type):
  """Attempt to find if the testcase affects release branches on Chromium."""
  # This shouldn't ever get scheduled, but check just in case.
  if not utils.is_chromium():
    return

  # Locate the testcase associated with the id.
  testcase = data_handler.get_testcase_by_id(testcase_id)

  # If this testcase is fixed, we should no longer be doing impact testing.
  if testcase.fixed and testcase.is_impact_set_flag:
    return

  # For testcases with status unreproducible, we just do impact analysis just
  # once.
  if testcase.is_status_unreproducible() and testcase.is_impact_set_flag:
    return

  # Update comments only after checking the above bailout conditions.
  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  # This task is not applicable to unreproducible testcases.
  if testcase.one_time_crasher_flag:
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.ERROR,
        'Not applicable for unreproducible testcases')
    return

  # This task is not applicable for custom binaries. We cannot remove the
  # creation of such tasks specifically for custom binary testcase in cron,
  # so exit gracefully.
  if build_manager.is_custom_binary():
    data_handler.update_testcase_comment(testcase,
                                         data_types.TaskState.FINISHED,
                                         'Not applicable for custom binaries')
    return

  # If we don't have a stable or beta build url pattern, we try to use build
  # information url to make a guess.
  if not build_manager.has_production_builds():
    if not testcase.regression:
      data_handler.update_testcase_comment(
          testcase, data_types.TaskState.FINISHED,
          'Cannot run without regression range, will re-run once regression '
          'task finishes')
      return

    impacts = get_impacts_from_url(testcase.regression, testcase.job_type)
    testcase = data_handler.get_testcase_by_id(testcase_id)
    set_testcase_with_impacts(testcase, impacts)
    data_handler.update_testcase_comment(testcase,
                                         data_types.TaskState.FINISHED)
    return

  # Setup testcase and its dependencies.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    return

  # Setup extended stable, stable, beta builds
  # and get impact and crash stacktrace.
  try:
    impacts = get_impacts_on_prod_builds(testcase, testcase_file_path)
  except BuildFailedException as error:
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         str(error))
    tasks.add_task(
        'impact',
        testcase_id,
        job_type,
        wait_time=environment.get_value('FAIL_WAIT'))
    return

  testcase = data_handler.get_testcase_by_id(testcase_id)
  set_testcase_with_impacts(testcase, impacts)

  # Set stacktrace in case we have a unreproducible crash on trunk,
  # but it crashes on one of the production builds.
  if testcase.is_status_unreproducible() and impacts.get_extra_trace():
    testcase.crash_stacktrace = data_handler.filter_stacktrace(
        '%s\n\n%s' % (data_handler.get_stacktrace(testcase),
                      impacts.get_extra_trace()))

  data_handler.update_testcase_comment(testcase, data_types.TaskState.FINISHED)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for AFL's engine implementation."""

import os
import shutil
import unittest

from clusterfuzz._internal.bot.fuzzers.afl import engine
from clusterfuzz._internal.bot.fuzzers.afl import launcher
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.core.bot.fuzzers.afl import \
    afl_launcher_integration_test
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

# TODO(mbarbella): Break dependency on afl_launcher_integration_test once
# everything has been fully converted to the new pipeline.

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
TEMP_DIRECTORY = os.path.join(TEST_PATH, 'temp')
DATA_DIRECTORY = os.path.join(TEST_PATH, 'data')
CORPUS_DIRECTORY = os.path.join(TEMP_DIRECTORY, 'corpus')
OUTPUT_DIRECTORY = os.path.join(TEMP_DIRECTORY, 'output')

BASE_FUZZ_TIMEOUT = (
    launcher.AflRunnerCommon.SIGTERM_WAIT_TIME +
    launcher.AflRunnerCommon.AFL_CLEAN_EXIT_TIME)
FUZZ_TIMEOUT = 5 + BASE_FUZZ_TIMEOUT
LONG_FUZZ_TIMEOUT = 90 + BASE_FUZZ_TIMEOUT


def clear_temp_dir():
  """Clear temp directories."""
  if os.path.exists(TEMP_DIRECTORY):
    shutil.rmtree(TEMP_DIRECTORY)


def create_temp_dir():
  """Create temp directories."""
  # Corpus directory will be created when preparing for fuzzing.
  os.mkdir(TEMP_DIRECTORY)
  os.mkdir(OUTPUT_DIRECTORY)


@unittest.skipIf(not environment.get_value('AFL_INTEGRATION_TESTS'),
                 'AFL_INTEGRATION_TESTS=1 must be set')
class AFLEngineTest(unittest.TestCase):
  """Tests for AFLEngine."""

  def setUp(self):
    clear_temp_dir()
    create_temp_dir()

    test_helpers.patch_environ(self)
    afl_launcher_integration_test.dont_use_strategies(self)

  def tearDown(self):
    clear_temp_dir()

  def test_fuzz(self):
    """Test for fuzz."""
    engine_impl = engine.AFLEngine()

    afl_launcher_integration_test.setup_testcase_and_corpus(
        'empty', 'corpus', fuzz=True)
    fuzzer_path = os.path.join(DATA_DIRECTORY, 'test_fuzzer')
    options = engine_impl.prepare(CORPUS_DIRECTORY, fuzzer_path, DATA_DIRECTORY)

    result = engine_impl.fuzz(fuzzer_path, options, OUTPUT_DIRECTORY,
                              FUZZ_TIMEOUT)

    self.assertEqual('{0}/afl-fuzz'.format(DATA_DIRECTORY), result.command[0])
    self.assertIn('-i{0}'.format(CORPUS_DIRECTORY), result.command)

    # Ensure that we've added something other than the dummy file to the corpus.
    self.assertTrue(os.listdir(CORPUS_DIRECTORY))

  def test_reproduce(self):
    """Test for reproduce."""
    engine_impl = engine.AFLEngine()
    target_path = os.path.join(DATA_DIRECTORY, 'test_fuzzer')
    testcase_path = afl_launcher_integration_test.setup_testcase_and_corpus(
        'crash', 'empty_corpus')
    timeout = 5
    result = engine_impl.reproduce(target_path, testcase_path, [], timeout)

    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000',
        result.output)

  def test_fuzz_with_crash(self):
    """Tests that we detect crashes when fuzzing."""
    engine_impl = engine.AFLEngine()

    afl_launcher_integration_test.setup_testcase_and_corpus(
        'empty', 'corpus', fuzz=True)
    fuzzer_path = os.path.join(DATA_DIRECTORY, 'easy_crash_fuzzer')
    options = engine_impl.prepare(CORPUS_DIRECTORY, fuzzer_path, DATA_DIRECTORY)

    result = engine_impl.fuzz(fuzzer_path, options, OUTPUT_DIRECTORY,
                              LONG_FUZZ_TIMEOUT)

    self.assertGreater(len(result.crashes), 0)
    crash = result.crashes[0]
    self.assertIn('ERROR: AddressSanitizer: heap-use-after-free',
                  crash.stacktrace)

    # Testcase (non-zero size) should've been copied back.
    self.assertNotEqual(os.path.getsize(crash.input_path), 0)

  def test_startup_crash_not_reported(self):
    """Ensures that we properly handle startup crashes."""
    engine_impl = engine.AFLEngine()

    afl_launcher_integration_test.setup_testcase_and_corpus(
        'empty', 'corpus', fuzz=True)
    fuzzer_path = os.path.join(DATA_DIRECTORY, 'always_crash_fuzzer')
    options = engine_impl.prepare(CORPUS_DIRECTORY, fuzzer_path, DATA_DIRECTORY)

    result = engine_impl.fuzz(fuzzer_path, options, OUTPUT_DIRECTORY,
                              FUZZ_TIMEOUT)

    self.assertFalse(result.crashes)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""logs test."""
import inspect
import json
import logging
import os
import sys
import unittest

import mock

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.tests.test_libs import helpers


class GetSourceLocationTest(unittest.TestCase):
  """Test get_source_location."""

  def _nested(self):
    self.statement_line = inspect.currentframe().f_lineno + 1
    return logs.get_source_location()

  def test_get(self):
    """Test get."""
    pathname, lineno, func_name = self._nested()
    self.assertTrue(
        pathname.endswith(
            'src/clusterfuzz/_internal/tests/core/metrics/logs_test.py'))
    # The line number of _dummy_emit() invocation.
    self.assertEqual(self.statement_line, lineno)
    self.assertEqual('_nested', func_name)


class UpdateEntryWithExc(unittest.TestCase):
  """Test update_entry_with_exc."""

  def test_empty(self):
    """Test empty exc_info."""
    entry = {}
    logs.update_entry_with_exc(entry, None)
    self.assertEqual({}, entry)

  def test_none_exc(self):
    """Test exc_info is (None, None, None)."""
    entry = {
        'task_payload': 'task',
        'extras': {
            'test': 'value'
        },
        'message': 'original',
        'serviceContext': {
            'service': 'bots'
        },
        'location': {
            'path': 'source_path',
            'line': 1234,
            'method': 'new_method'
        }
    }
    exc_info = sys.exc_info()  # expected to be (None, None, None).

    logs.update_entry_with_exc(entry, exc_info)

    self.assertEqual({
        'extras': {
            'test': 'value'
        },
        'task_payload': 'task',
        'message': 'original',
        'serviceContext': {
            'service': 'bots'
        },
        'location': {
            'path': 'source_path',
            'line': 1234,
            'method': 'new_method'
        },
        'context': {
            'reportLocation': {
                'filePath': 'source_path',
                'lineNumber': 1234,
                'functionName': 'new_method'
            }
        }
    }, entry)

  def test_exc(self):
    """Test exception."""
    entry = {'extras': {}, 'message': 'original'}
    exception = Exception('ex message')
    exception.extras = {'test': 'value', 'task_payload': 'task'}

    try:
      statement_line = inspect.currentframe().f_lineno + 1
      raise exception
    except:
      # We do this because we need the traceback instance.
      exc_info = sys.exc_info()

    logs.update_entry_with_exc(entry, exc_info)
    self.maxDiff = None  # pylint: disable=invalid-name

    self.assertEqual({
        'extras': {
            'test': 'value'
        },
        'task_payload':
            'task',
        'serviceContext': {
            'service': 'bots'
        },
        'message':
            ('original\n'
             'Traceback (most recent call last):\n'
             '  File "%s", line %d, in test_exc\n'
             '    raise exception\n'
             'Exception: ex message\n') % (__file__.strip('c'), statement_line)
    }, entry)


class FormatRecordTest(unittest.TestCase):
  """Test format_record."""

  def setUp(self):
    helpers.patch(self,
                  ['clusterfuzz._internal.metrics.logs.update_entry_with_exc'])
    helpers.patch_environ(self)

    self.maxDiff = None  # pylint: disable=invalid-name

  def get_record(self):
    """Make a fake record."""
    os.environ['BOT_NAME'] = 'linux-bot'
    os.environ['TASK_PAYLOAD'] = 'fuzz fuzzer1 job1'
    record = mock.Mock(
        specset=logging.LogRecord,
        levelname='INFO',
        exc_info='exc_info',
        created=10,
        location={
            'path': 'path',
            'line': 123,
            'method': 'func'
        })
    record.name = 'logger_name'
    record.getMessage.return_value = 'log message'
    return record

  def test_format_record(self):
    """Test format a LogRecord into JSON string."""
    os.environ['FUZZ_TARGET'] = 'fuzz_target1'
    record = self.get_record()
    record.extras = {'a': 1}
    self.assertEqual({
        'message': 'log message',
        'created': '1970-01-01T00:00:10Z',
        'severity': 'INFO',
        'bot_name': 'linux-bot',
        'task_payload': 'fuzz fuzzer1 job1',
        'fuzz_target': 'fuzz_target1',
        'name': 'logger_name',
        'extras': {
            'a': 1,
        },
        'location': {
            'path': 'path',
            'line': 123,
            'method': 'func'
        }
    }, json.loads(logs.format_record(record)))

    self.mock.update_entry_with_exc.assert_called_once_with(
        mock.ANY, 'exc_info')

  def test_no_extras(self):
    """Test format record with no extras."""
    record = self.get_record()
    record.extras = None
    self.assertEqual({
        'message': 'log message',
        'created': '1970-01-01T00:00:10Z',
        'severity': 'INFO',
        'bot_name': 'linux-bot',
        'task_payload': 'fuzz fuzzer1 job1',
        'name': 'logger_name',
        'location': {
            'path': 'path',
            'line': 123,
            'method': 'func'
        }
    }, json.loads(logs.format_record(record)))
    self.mock.update_entry_with_exc.assert_called_once_with(
        mock.ANY, 'exc_info')

  def test_worker_bot_name(self):
    """Test format record with a worker bot name."""
    os.environ['WORKER_BOT_NAME'] = 'worker'
    record = self.get_record()
    record.extras = None

    self.assertEqual({
        'message': 'log message',
        'created': '1970-01-01T00:00:10Z',
        'severity': 'INFO',
        'bot_name': 'linux-bot',
        'worker_bot_name': 'worker',
        'task_payload': 'fuzz fuzzer1 job1',
        'name': 'logger_name',
        'location': {
            'path': 'path',
            'line': 123,
            'method': 'func'
        }
    }, json.loads(logs.format_record(record)))
    self.mock.update_entry_with_exc.assert_called_once_with(
        mock.ANY, 'exc_info')


class JsonSocketHandler(unittest.TestCase):
  """Test JsonSocketHandler."""

  def setUp(self):
    helpers.patch(self, ['clusterfuzz._internal.metrics.logs.format_record'])

  def test_make_pickle(self):
    """Test makePickle."""
    self.mock.format_record.return_value = 'json'

    record = mock.Mock()
    handler = logs.JsonSocketHandler(None, None)
    self.assertEqual(b'json\n', handler.makePickle(record))

    self.mock.format_record.assert_called_once_with(record)


class ConfigureTest(unittest.TestCase):
  """Test configure."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.get_logging_config_dict',
        'clusterfuzz._internal.metrics.logs.set_logger',
        'logging.config.dictConfig',
        'logging.getLogger',
        'clusterfuzz._internal.metrics.logs._is_running_on_app_engine',
        'clusterfuzz._internal.metrics.logs.suppress_unwanted_warnings',
        'google.cloud.logging.Client',
    ])

  def test_configure(self):
    """Test configure."""
    self.mock._is_running_on_app_engine.return_value = False  # pylint: disable=protected-access
    logs._logger = None  # pylint: disable=protected-access
    logger = mock.MagicMock()
    self.mock.getLogger.return_value = logger

    logs.configure('test')

    self.mock.set_logger.assert_called_with(logger)
    self.mock.get_logging_config_dict.assert_called_once_with('test')
    self.mock.getLogger.assert_called_with('test')
    self.mock.dictConfig.assert_called_once_with(
        self.mock.get_logging_config_dict.return_value)

  def test_configure_appengine(self):
    """Test configure on App Engine."""
    self.mock._is_running_on_app_engine.return_value = True  # pylint: disable=protected-access
    logs.configure('test')
    self.assertEqual(0, self.mock.dictConfig.call_count)


class EmitTest(unittest.TestCase):
  """Test emit."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.get_logger',
        'clusterfuzz._internal.metrics.logs._is_running_on_app_engine'
    ])
    # Reset default extras as it may be modified during other test runs.
    logs._default_extras = {}  # pylint: disable=protected-access
    self.mock._is_running_on_app_engine.return_value = False  # pylint: disable=protected-access

  def test_no_logger(self):
    """Test no logger."""
    self.mock.get_logger.return_value = None
    logs.emit(logging.INFO, 'message')

  def test_log_info(self):
    """Test log info."""
    logger = mock.MagicMock()
    self.mock.get_logger.return_value = logger

    statement_line = inspect.currentframe().f_lineno + 1
    logs.emit(logging.INFO, 'msg', target='bot', test='yes')

    logger.log.assert_called_once_with(
        logging.INFO,
        'msg',
        exc_info=None,
        extra={
            'extras': {
                'target': 'bot',
                'test': 'yes'
            },
            'location': {
                'path': os.path.abspath(__file__).rstrip('c'),
                'line': statement_line,
                'method': 'test_log_info'
            }
        })

  def test_log_error(self):
    """Test log error."""
    logger = mock.MagicMock()
    self.mock.get_logger.return_value = logger

    statement_line = inspect.currentframe().f_lineno + 1
    logs.emit(logging.ERROR, 'msg', exc_info='ex', target='bot', test='yes')

    logger.log.assert_called_once_with(
        logging.ERROR,
        'msg',
        exc_info='ex',
        extra={
            'extras': {
                'target': 'bot',
                'test': 'yes'
            },
            'location': {
                'path': os.path.abspath(__file__).rstrip('c'),
                'line': statement_line,
                'method': 'test_log_error'
            }
        })


class TruncateTest(unittest.TestCase):
  """Test truncate."""

  def test_not_truncate(self):
    """Test msg is not too long."""
    self.assertEqual('abcd', logs.truncate('abcd', 4))

  def test_truncate(self):
    """Test truncate because msh is too long."""
    self.assertEqual('abc\n...5 characters truncated...\nijk',
                     logs.truncate('abcdefghijk', 6))


class LogErrorTest(unittest.TestCase):
  """Tests log_error."""

  def setUp(self):
    helpers.patch(self,
                  ['clusterfuzz._internal.metrics.logs.emit', 'sys.exc_info'])

  def test_no_exception(self):
    """Tests no exception."""
    self.mock.exc_info.return_value = 'err'
    logs.log_error('test', hello='1')
    self.mock.emit.assert_called_once_with(
        logging.ERROR, 'test', exc_info='err', hello='1')

  def test_exception(self):
    """Tests exception."""
    self.mock.exc_info.return_value = 'err'
    logs.log_error('test', exception='exception', hello='1')
    self.mock.emit.assert_called_once_with(
        logging.ERROR, 'test', exc_info='err', hello='1')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for bots."""
import collections
import datetime
import string
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import bots


@test_utils.with_cloud_emulators('datastore')
class BotsTest(unittest.TestCase):
  """Jobs tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.has_access',
        'libs.access.get_access',
        'libs.gcs.prepare_blob_upload',
    ])
    self.mock.prepare_blob_upload.return_value = (
        collections.namedtuple('GcsUpload', [])())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=bots.JsonHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def _create_bot(self,
                  bot_name,
                  last_beat_time=datetime.datetime(2020, 1, 25, 0, 0),
                  task_payload='',
                  task_end_time=datetime.datetime(2020, 1, 25, 0, 0),
                  source_version='',
                  platform_id=''):
    """Create a test job."""
    bot = data_types.Heartbeat()
    bot.bot_name = bot_name
    bot.last_beat_time = last_beat_time
    bot.task_payload = task_payload
    bot.task_end_time = task_end_time
    bot.source_version = source_version
    bot.platform_id = platform_id
    bot.put()

    return bot

  def test_pagination(self):
    """Test bots pagination and post method."""
    self.mock.has_access.return_value = True
    expected_items = {1: [], 2: [], 3: []}

    for bot_num, bot_suffix in enumerate(string.ascii_lowercase):
      bot_name = "test_bot_" + bot_suffix
      bot = self._create_bot(bot_name=bot_name)
      expected_items[(bot_num // bots.PAGE_SIZE) + 1].append(bot.bot_name)

    resp = self.app.post_json('/', {'page': 1})
    self.assertListEqual(expected_items[1],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 2})
    self.assertListEqual(expected_items[2],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 3})
    self.assertListEqual(expected_items[3],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'page': 4})
    self.assertListEqual([], [item['bot_name'] for item in resp.json['items']])

  def test_search(self):
    """Test bots search."""
    self.mock.has_access.return_value = True

    bot_a = self._create_bot(bot_name='test_bot_a', task_payload='pay_x')
    bot_b = self._create_bot(bot_name='test_bot_b', task_payload='pay_y')

    resp = self.app.post_json('/', {'q': 'a'})
    self.assertListEqual([bot_a.bot_name],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'b'})
    self.assertListEqual([bot_b.bot_name],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'c'})
    self.assertListEqual([], [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'bot'})
    self.assertListEqual([bot_a.bot_name, bot_b.bot_name],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'x'})
    self.assertListEqual([bot_a.bot_name],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'y'})
    self.assertListEqual([bot_b.bot_name],
                         [item['bot_name'] for item in resp.json['items']])

    resp = self.app.post_json('/', {'q': 'pay'})
    self.assertListEqual([bot_a.bot_name, bot_b.bot_name],
                         [item['bot_name'] for item in resp.json['items']])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuchsia utilities for handling fuzzers."""

import datetime
import errno
import glob
import os
import subprocess
import time

import six


class Fuzzer(object):
  """Represents a Fuchsia fuzz target.

    This represents a binary fuzz target produced the Fuchsia build, referenced
    by a component manifest, and included in a fuzz package.  It provides an
    interface for running the fuzzer in different common modes, allowing
    specific command line arguments to libFuzzer to be abstracted.

    Attributes:
      device: A Device where this fuzzer can be run
      host: The build host that built the fuzzer
      pkg: The GN fuzzers_package name
      tgt: The GN fuzzers name
  """

  # Matches the prefixes in libFuzzer passed to |Fuzzer::DumpCurrentUnit| or
  # |Fuzzer::WriteUnitToFileWithPrefix|.
  ARTIFACT_PREFIXES = [
      'crash', 'leak', 'mismatch', 'oom', 'slow-unit', 'timeout'
  ]

  class NameError(ValueError):
    """Indicates a supplied name is malformed or unusable."""

  class StateError(ValueError):
    """Indicates a command isn't valid for the fuzzer in its current state."""

  @classmethod
  def _matches_sanitizer(cls, tgt, sanitizer):
    """ Returns whether or not a target is for a given sanitizer. """
    tgt, ext = os.path.splitext(tgt)
    # Remove the leading '.' from the extension.
    ext = ext[1:]
    if ext == sanitizer:
      return True
    # If there's no extension, assume it's an ASAN fuzzer.
    if sanitizer == 'asan' and ext == '':
      return True
    return False

  @classmethod
  def _is_example(cls, pkg, tgt):
    """ Returns whether or not a given pkg/tgt pair is an example fuzzer.
    (Helper function to prevent us from wasting cycles on example fuzzers in
    production). """
    # Strip any sanitizer extensions
    tgt = os.path.splitext(tgt)[0]
    return ((pkg == 'example-fuzzers' and
             tgt not in ('out_of_memory_fuzzer', 'toy_example_arbitrary')) or
            (pkg == 'zircon_fuzzers' and tgt == 'noop-fuzzer'))

  @classmethod
  def filter(cls, fuzzers, name, sanitizer=None, example_fuzzers=True):
    """Filters a list of fuzzer names.

      Takes a list of fuzzer names in the form `pkg`/`tgt` and a name to filter
      on.  If the name is of the form 'x/y', the filtered list will include all
      the fuzzer names where 'x' is a substring of `pkg` and y is a substring
      of `tgt`; otherwise it includes all the fuzzer names where `name` is a
      substring of either `pkg` or `tgt`.

      Returns:
        A list of fuzzer names matching the given name.

      Raises:
        FuzzerNameError: Name is malformed, e.g. of the form 'x/y/z'.
    """
    if not name and not sanitizer:
      return fuzzers
    if name:
      names = name.split('/')
      if len(names) == 2 and (names[0], names[1]) in fuzzers:
        return [(names[0], names[1])]
      if len(names) == 1:
        return list(
            set(Fuzzer.filter(fuzzers, '/' + name))
            | set(Fuzzer.filter(fuzzers, name + '/')))
      if len(names) != 2:
        raise Fuzzer.NameError('Malformed fuzzer name: ' + name)
    filtered = []
    for pkg, tgt in fuzzers:
      if name:
        if not (names[0] in pkg and names[1] in tgt):
          continue
      if sanitizer:
        if not Fuzzer._matches_sanitizer(tgt, sanitizer):
          continue
      if not example_fuzzers:
        if Fuzzer._is_example(pkg, tgt):
          continue
      # Remove the sanitizer extension name.
      # Clusterfuzz only needs to know foo/bar, not foo/bar.asan.
      filtered.append((pkg, os.path.splitext(tgt)[0]))
    return filtered

  @classmethod
  def from_args(cls, device, args):
    """Constructs a Fuzzer from command line arguments."""
    fuzzers = Fuzzer.filter(device.host.fuzzers, args.name)
    if len(fuzzers) != 1:
      raise Fuzzer.NameError('Name did not resolve to exactly one fuzzer: \'' +
                             args.name + '\'. Try using \'list-fuzzers\'.')
    return cls(device, fuzzers[0][0], fuzzers[0][1], args.output,
               args.foreground)

  def __init__(self,
               device,
               pkg,
               tgt,
               output=None,
               foreground=False,
               sanitizer=''):
    self.device = device
    self.host = device.host
    self.pkg = pkg
    self.tgt = tgt
    self.last_fuzz_cmd = None
    if output:
      self._output = output
    else:
      self._output = self.host.join('test_data', 'fuzzing', self.pkg, self.tgt)
    self._results_output = self.host.join('test_data', 'fuzzing', self.pkg,
                                          self.tgt)
    self._foreground = foreground

    # Required for backwards compatibility with older builds where Zircon
    # fuzzers had a sanitizer suffix
    if pkg == 'zircon_fuzzers' and sanitizer:
      if (pkg, tgt) not in self.host.fuzzers:
        self.tgt += '.' + sanitizer

  def __str__(self):
    return self.pkg + '/' + self.tgt

  def data_path(self, relpath=''):
    """Canonicalizes the location of mutable data for this fuzzer."""
    return '/data/r/sys/fuchsia.com:%s:0#meta:%s.cmx/%s' % (self.pkg, self.tgt,
                                                            relpath)

  def measure_corpus(self):
    """Returns the number of corpus elements and corpus size as a pair."""
    try:
      sizes = self.device.ls(self.data_path('corpus'))
      return (len(sizes), sum(sizes.values()))
    except subprocess.CalledProcessError:
      return (0, 0)

  def list_artifacts(self):
    """Returns a list of test unit artifacts, i.e. fuzzing crashes."""
    artifacts = []
    try:
      lines = self.device.ls(self.data_path())
      for artifact, _ in six.iteritems(lines):
        for prefix in Fuzzer.ARTIFACT_PREFIXES:
          if artifact.startswith(prefix):
            artifacts.append(artifact)
      return artifacts
    except subprocess.CalledProcessError:
      return []

  def is_running(self):
    """Checks the device and returns whether the fuzzer is running."""
    return self.tgt in self.device.getpids()

  def require_stopped(self):
    """Raise an exception if the fuzzer is running."""
    if self.is_running():
      raise Fuzzer.StateError(
          str(self) + ' is running and must be stopped first.')

  def results(self, relpath=None):
    """Returns the path in the previously prepared results directory."""
    if relpath:
      return os.path.join(self._output, 'latest', relpath)
    return os.path.join(self._output, 'latest')

  def results_output(self, relpath=None):
    if relpath:
      return os.path.join(self._results_output, relpath)
    return self._results_output

  def url(self):
    return 'fuchsia-pkg://fuchsia.com/%s#meta/%s.cmx' % (self.pkg, self.tgt)

  def run(self, fuzzer_args, logfile=None):
    fuzz_cmd = ['run', self.url(), '-artifact_prefix=data/'] + fuzzer_args
    print('+ ' + ' '.join(fuzz_cmd))
    self.last_fuzz_cmd = self.device.get_ssh_cmd(['ssh', 'localhost'] +
                                                 fuzz_cmd)
    return self.device.ssh(fuzz_cmd, quiet=True, logfile=logfile)

  def start(self, fuzzer_args):
    """Runs the fuzzer.

      Executes a fuzzer in the "normal" fuzzing mode. It spawns the fuzzer,
      but does not wait until it completes. As a result, callers will
      typically want to subsequently call Fuzzer.monitor()

      The command will be like:
      run fuchsia-pkg://fuchsia.com/<pkg>#meta/<tgt>.cmx \
        -artifact_prefix=data/ -jobs=1 data/corpus/

      See also: https://llvm.org/docs/LibFuzzer.html#running

      Args:
        fuzzer_args: Command line arguments to pass to libFuzzer

      Returns:
        The fuzzer's process ID. May be 0 if the fuzzer stops immediately.
    """
    self.require_stopped()
    results = os.path.join(self._output, datetime.datetime.utcnow().isoformat())
    try:
      os.makedirs(results)
    except OSError as e:
      if e.errno != errno.EEXIST:
        raise
    self._results_output = results
    self.logfile = self.results_output('fuzz-0.log')

    if [x for x in fuzzer_args if x.startswith('-jobs=')]:
      if self._foreground:
        fuzzer_args.append('-jobs=0')
      else:
        fuzzer_args.append('-jobs=1')
    self.device.ssh(['mkdir', '-p', self.data_path('corpus')])
    # If all the arguments are prepended with '-', then no corpus has been
    # passed in, and we need to add one.
    # This list comprehension returns a list of all arguments that do *not*
    # start with '-'.
    # Thus, if the list is empty, we append data/corpus/.
    # TODO(flowerhack): Strictly speaking, libfuzzer doesn't *need* a corpus
    # directory to run, and a user may find it confusing that one is
    # automagically created.
    # Change this to *not* be the default.  (If we'd like it to make it the
    # default for e.g. most non-Clusterfuzz callers, we can have some variable
    # that those callers pass in to set this.)
    if not [x for x in fuzzer_args if not x.startswith('-')]:
      fuzzer_args.append('data/corpus/')
    if 'repro' in fuzzer_args:
      # If this is a reproducer run, we don't need the corpus.
      if 'data/corpus/' in fuzzer_args:
        fuzzer_args.remove('data/corpus/')
      fuzzer_args.remove('repro')

    # Fuzzer logs are saved to fuzz-*.log when running in the background.
    # We tee the output to fuzz-0.log when running in the foreground to
    # make the rest of the plumbing look the same.
    if self._foreground:
      return self.run(fuzzer_args, logfile=self.results_output('fuzz-0.log'))
    self.device.rm(self.data_path('fuzz-*.log'))
    return self.run(fuzzer_args)

  def monitor(self, retcode=0):
    """Waits for a fuzzer to complete and symbolizes its logs.

        Polls the device to determine when the fuzzer stops. Retrieves,
        combines and symbolizes the associated fuzzer and kernel logs. Fetches
        any referenced test artifacts, e.g. crashes.
        """
    while self.is_running():
      time.sleep(2)
    if not self._foreground:
      self.device.fetch(self.data_path('fuzz-*.log'), self.results_output())
    logs = glob.glob(self.results_output('fuzz-*.log'))
    guess_pid = len(logs) == 1
    artifacts = []
    for log in logs:
      artifacts += self.device.process_logs(log, guess_pid, retcode)
    for artifact in artifacts:
      self.device.fetch(self.data_path(artifact), self.results_output())

  def stop(self):
    """Stops any processes with a matching component manifest on the device."""
    pids = self.device.getpids()
    if self.tgt in pids:
      self.device.ssh(['kill', str(pids[self.tgt])])

  def repro(self, fuzzer_args):
    """Runs the fuzzer with test input artifacts.

      Executes a command like:
      run fuchsia-pkg://fuchsia.com/<pkg>#meta/<tgt>.cmx \
        -artifact_prefix=data -jobs=1 data/<artifact>...

      See also: https://llvm.org/docs/LibFuzzer.html#options

      Returns: Number of test input artifacts found.
    """
    artifacts = self.list_artifacts()
    if artifacts:
      self.run(fuzzer_args + ['data/' + a for a in artifacts])
    return len(artifacts)

  def merge(self, fuzzer_args, merge_control_file=None):
    """Attempts to minimizes the fuzzer's corpus.

      Executes a command like:
      run fuchsia-pkg://fuchsia.com/<pkg>#meta/<tgt>.cmx \
        -artifact_prefix=data -jobs=1 \
        -merge=1 -merge_control_file=data/.mergefile \
        data/corpus/ data/corpus.prev/'

      See also: https://llvm.org/docs/LibFuzzer.html#corpus

      Returns: Same as measure_corpus
    """
    self.require_stopped()
    if self.measure_corpus() == (0, 0):
      return (0, 0)

    # If all the arguments are prepended with '-', then no corpus has been
    # passed in, so we need to execute the "default" behavior: assuming
    # "corpus" is the only relevant directory, and making a new corpus
    # directory.
    # If corpora have been passed in, we trust that the caller has passed
    # them in the order they want.
    if not [x for x in fuzzer_args if not x.startswith('-')]:
      self.device.ssh(['mkdir', '-p', self.data_path('corpus')])
      self.device.ssh(['mkdir', '-p', self.data_path('corpus.prev')])
      self.device.ssh(
          ['mv',
           self.data_path('corpus/*'),
           self.data_path('corpus.prev')])
      self.device.ssh(['mkdir', '-p', self.data_path('corpus')])
      fuzzer_args.append('data/corpus/')
      fuzzer_args.append('data/corpus.prev/')

    if not merge_control_file:
      merge_control_file = 'data/.mergefile'

    merge_args = ['-merge=1', '-merge_control_file=' + merge_control_file]
    self.run(merge_args + fuzzer_args)

    # Cleanup
    self.device.rm(self.data_path('.mergefile'))
    self.device.rm(self.data_path('corpus.prev'), recursive=True)
    return self.measure_corpus()
